"""
used for logging
"""

import os
from collections import namedtuple
from datetime import datetime
from typing import Any, Callable

from ..utils.dates import curent_date
from ..utils.screen import ScreenStr

loginfo = namedtuple('loginfo', ['string', 'prefix_len'])


class Logger:
    VERBOSE = 0
    V_DEBUG = -1
    V_INFO = 0
    V_WARN = 1
    V_ERROR = 2
    V_FATAL = 3
    _instance = None

    def __new__(cls, *args, **kwargs) -> Any:
        if Logger._instance is not None:
            return Logger._instance
        return super().__new__(cls)

    def __init__(self, adddate=True, datefmt: str = '%y-%m-%d %H:%M:%S', sep: str = " | ", use_stdout: bool = True):
        if Logger._instance is not None:
            return

        self.adddate = adddate
        self.datefmt = datefmt
        self.out_channel = []
        self.pipe_key = set()
        self.sep = sep
        self.return_str = ""
        self.listener = []
        self.use_stdout = use_stdout
        Logger._instance = self

    def _format(self, *values, inline=False, fix=0, raw=False, append=False):
        """"""
        if self.adddate and not raw:
            cur_date = datetime.now().strftime(self.datefmt)
        else:
            cur_date = ""

        values = ["{}".format(str(i)) for i in values]
        values = [i for i in values if len(i.strip()) != 0]

        if len(cur_date) == 0:
            space = [*["{}".format(str(i)) for i in values]]
        else:
            space = [cur_date, *values]

        if fix >= 0:
            left, right = self.sep.join(space[:fix + 1]), self.sep.join(space[fix + 1:])
            fix = len(left) + len(self.sep)
            logstr = self.sep.join((left, right))

            if inline:
                if append:
                    return "{}".format(logstr), fix
                else:
                    return "\r{}".format(logstr), fix
            else:
                return "{}\n".format(logstr), fix

        space = self.sep.join(space)

        if inline:
            return loginfo("\r{}".format(space), 0)
        else:
            return loginfo("{}\n".format(space), 0)

    def inline(self, *values, fix=0, append=False):
        """Log a message with severity 'INFO' inline"""
        logstr, fix = self._format(*values, inline=True, fix=fix, append=append)
        self.handle(logstr, fix=fix)

    def info(self, *values):
        """Log a message with severity 'INFO'"""
        logstr, fix = self._format(*values, inline=False)
        self.handle(logstr, level=Logger.V_INFO, fix=fix)

    def raw(self, *values, inline=False, fix=0, level=0, append=False):
        """Log a message with severity 'INFO' withou datetime prefix"""
        logstr, fix = self._format(*values, inline=inline, fix=fix, raw=True, append=append)
        self.handle(logstr, level=level)

    def debug(self, *values):
        """Log a message with severity 'DEBUG'"""
        # TODO 添加红色前景
        logstr, fix = self._format("DEBUG", *values, inline=False)
        self.handle(logstr, level=Logger.V_DEBUG, fix=fix)

    def warn(self, *values):
        """Log a message with severity 'WARN'"""
        # TODO 添加黄色前景
        logstr, fix = self._format("WARN", *values, inline=False)
        self.handle(logstr, level=Logger.V_WARN, fix=fix)

    def error(self, *values):
        """Log a message with severity 'ERROR'"""
        # TODO 添加黄色背景
        logstr, fix = self._format("ERROR", *values, inline=False)
        self.handle(logstr, level=Logger.V_ERROR, fix=fix)

    def fatal(self, *values):
        """Log a message with severity 'FATAL'"""
        # TODO 添加红色背景
        logstr, fix = self._format("FATAL", *values, inline=False)
        self.handle(logstr, level=Logger.V_FATAL, fix=fix)

    def newline(self):
        """"""
        self.handle("")

    def handle(self, logstr, end="", level=0, **kwargs):
        """handle log stinrg"""
        for listener in self.listener:
            listener(logstr, end, level)

        if level < Logger.VERBOSE:
            return

        if logstr.startswith("\r"):
            fix = kwargs.get("fix", 0)
            self.return_str = logstr
            self.print(ScreenStr(logstr, leftoffset=fix), end=end)
        else:
            if len(self.return_str) != 0:
                self.print(self.return_str, end="\n")
            self.print(logstr, end="")

            for i in self.out_channel:
                with open(i, "a", encoding="utf-8") as w:
                    if len(self.return_str) != 0:
                        w.write("{}\n".format(self.return_str.strip()))
                    w.write(logstr)

            self.return_str = ""

    def print(self, *args, end='\n'):
        """built-in print function"""
        if self.use_stdout:
            print(*args, end=end, flush=True)

    def toggle_stdout(self, val:bool):
        """"""
        self.use_stdout = val

    def add_log_dir(self, dir):
        """add a file output pipeline"""
        if dir in self.pipe_key:
            self.info("Add pipe {}, but already exists".format(dir))
            return None

        os.makedirs(dir, exist_ok=True)

        i = 0
        cur_date = curent_date(fmt="%y%m%d%H%M%S")
        fni = os.path.join(dir, "l.{}.{}.log".format(cur_date, i))
        while os.path.exists(fni):
            i += 1
            fni = os.path.join(dir, "l.{}.{}.log".format(cur_date, i))

        self.print("add output channel on {}".format(fni))
        self.out_channel.append(fni)
        self.pipe_key.add(dir)
        return fni

    def add_log_listener(self, func: Callable[[str, str, int], Any]):
        """add a log event handler"""
        self.listener.append(func)

    @staticmethod
    def set_verbose(verbose=0):
        """"""
        Logger.VERBOSE = verbose



"""

"""

import os, sys
from functools import wraps

from ..calculate.schedule import Schedule, ScheduleList
from .meter import AvgMeter
from .meter import Meter
from .params import Params
from .trainer import Trainer
from ..base_classes.trickitems import NoneItem, AvgItem
from ..globals import _ML
from ..utils.timing import format_second


class BaseCallback():
    """
    base callback class

    only have two methods `on_begin()` and `on_end()`.

    for simpler using, see TrainCallback.
    """
    priority = 0  # type:int # All callbacks in thexp will have priority in range 0-100
    only_single_gpu = False  # only hooked in single gpu mode
    only_main_process = False  # whether can be hooked in children process( local_rank > 0)

    def __new__(cls, *_, **__):
        self = super().__new__(cls)
        self._trainer = None

        def ecp_wrap(func):
            """同一个异常第一次调用的时候运行"""

            @wraps(func)
            def on_exception(trainer: Trainer, tfunc, params: Params, e: BaseException, *args, **kwargs):
                self.ecp = getattr(self, "ecp", None)
                res = None
                if self.ecp != e:
                    res = self.on_first_exception(trainer, tfunc, params, e, *args, **kwargs)
                    self.ecp = e

                eres = func(trainer, tfunc, params, e, *args, **kwargs)
                if res is None:
                    return eres
                else:
                    return res

            return on_exception

        self.on_exception = ecp_wrap(self.on_exception)
        return self

    def on_hooked(self, trainer: Trainer, params: Params):
        """called when callback hooked trainer"""
        pass

    def on_first_exception(self, trainer: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        """
        when an exception was raised in some function, on_exception() will be called.

        如果异常发生在一个嵌套调用的函数中，那么该异常会在每一层函数都raise一次。

        该方法将被调用当该异常第一次raise出来的时候。
        该方法在 __new__ 中作了处理逻辑，不受继承关系影响
        """
        pass

    def on_exception(self, trainer: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        """called when exception raised in some function"""
        return False

    def on_hook_failed(self, trainer, message):
        """Any reason when callback cannot hook on trainer"""
        pass

    def on_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        """called before trainer.func is called"""
        pass

    def on_end(self, trainer: Trainer, func, params: Params, meter, *args, **kwargs):
        pass

    def __le__(self, other):
        return self.priority <= other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def hook(self, trainer: Trainer):
        """自动将自己已有的on_func_begin/on_func_end方法绑定"""
        # trainer.add_callback(self)
        if self.only_main_process and trainer.params.local_rank > 0:
            pass
        elif self.only_single_gpu and trainer.params.distributed:
            pass
        else:
            trainer.reload_callback(self)

    def unhook(self):
        self._trainer.remove_callback(self)

    def _repr_by_val(self, *vals):
        vstr = "; ".join(["{}={}".format(val, str(getattr(self, val, None))) for val in vals])
        return "<{}([{}]) at 0x{:016X}>".format(self.__class__.__name__, vstr, id(self))

    def __repr__(self) -> str:
        return self._repr_by_val("priority")


class TrainCallback(BaseCallback):
    """
    实现了一般训练过程中的函数函数回调，主要将 on_begin() / on_end() 方法分发到各具体的回调方法中
    """

    def on_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        if func.__name__ == "train":
            self.on_train_begin(trainer, func, params, *args, **kwargs)
        elif func.__name__ == "train_epoch":
            self.on_train_epoch_begin(trainer, func, params, *args, **kwargs)
        elif func.__name__ == "train_batch":
            self.on_train_batch_begin(trainer, func, params, *args, **kwargs)
        elif func.__name__ == "test":
            self.on_test_begin(trainer, func, params, *args, **kwargs)
        elif func.__name__ == "eval":
            self.on_eval_begin(trainer, func, params, *args, **kwargs)
        # elif func.__name__ in ""

    def on_initial_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_train_epoch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_test_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_eval_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_train_batch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_end(self, trainer: Trainer, func, params: Params, meter, *args, **kwargs):
        if func.__name__ == "train":
            self.on_train_end(trainer, func, params, meter, *args, **kwargs)
        elif func.__name__ == "train_epoch":
            self.on_train_epoch_end(trainer, func, params, meter, *args, **kwargs)
        elif func.__name__ == "train_batch":
            self.on_train_batch_end(trainer, func, params, meter, *args, **kwargs)
        elif func.__name__ == "test":
            self.on_test_end(trainer, func, params, meter, *args, **kwargs)
        elif func.__name__ == "eval":
            self.on_eval_end(trainer, func, params, meter, *args, **kwargs)
        elif func.__name__ == 'initial':
            self.on_initial_end(trainer, func, params, meter, *args, **kwargs)

    def on_train_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_test_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_eval_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_train_batch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass


class EvalCallback(TrainCallback):
    """
    决定在训练过程中，eval 和 test 的频率，当 Trainer 中没有注册训练集或测试集时，相应的过程会被跳过
    """
    only_main_process = True

    def __init__(self, eval_per_epoch=1, test_per_epoch=10):
        self.eval_in_per_epoch = eval_per_epoch
        self.test_in_per_epoch = test_per_epoch

        # 在训练结束后会进行一次
        self._last_eval = -1
        self._last_test = -1

    def _test_or_eval(self, params: Params, trainer: Trainer):
        if self.eval_in_per_epoch is not None and self.eval_in_per_epoch > 0:
            if params.eidx % self.eval_in_per_epoch == self.eval_in_per_epoch - 1:
                self._last_eval = params.eidx
                trainer.eval()
        if self.test_in_per_epoch is not None and self.test_in_per_epoch > 0:
            if params.eidx % self.test_in_per_epoch == self.test_in_per_epoch - 1:
                self._last_test = params.eidx
                trainer.test()

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self._test_or_eval(params, trainer)

    def on_train_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if self._last_eval != params.eidx:
            trainer.eval()
        if self._last_test != params.eidx:
            trainer.test()

    def __repr__(self):
        return self._repr_by_val("eval_in_per_epoch", "test_per_epoch")


class LoggerCallback(TrainCallback):
    """
    用于日志输出的回调，当 Trainer 在 epoch / batch 等级别的训练结束、异常发生等过程后，Logger 会对这些事件，
    或方法返回的结果进行输出。

    一般情况下 Logger 支持所有类型输出，但如果使用 Meter 类进行包装，会有更好的输出形式
    """
    only_main_process = True
    priority = 100

    def __init__(self, avg=True):
        self.avg = avg

    def on_hooked(self, trainer: Trainer, params: Params):
        super().on_hooked(trainer, params)
        trainer.logger.raw(' '.join(sys.argv))
        trainer.logger.info("Exp BaseDir", os.path.abspath(trainer.experiment.exp_dir))
        trainer.logger.info("Exp Trainer", trainer.__class__.__name__)
        trainer.logger.info("Exp Params")
        trainer.logger.raw(params)
        self.start = 0
        self.cur = None

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        from ..utils.timing import TimeIt

        self.start = params.eidx
        self.traintime = TimeIt()
        self.traintime.start()
        trainer.logger.info(trainer._databundler_dict)
        super().on_train_begin(trainer, func, params, *args, **kwargs)

    def on_train_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.traintime.end()
        if meter is None:
            meter = ""
        trainer.logger.info("train end", meter)
        trainer.logger.info("train time: {}".format(format_second(self.traintime["use"])))

    def on_train_epoch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        from ..utils.timing import TimeIt

        if self.avg:
            self.meter = AvgMeter()
        self.epochtime = TimeIt()
        self.epochtime.start()
        trainer.logger.info("{}/{}".format(params.eidx, params.epoch))

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.traintime.mark("epoch")
        self.epochtime.end()
        if self.cur is None:
            self.cur = params.eidx

        avg = self.traintime["use"] / (self.cur - self.start + 1)
        self.cur += 1
        last = (params.epoch - params.eidx) * avg

        tm = Meter()
        tm.train = format_second(self.traintime["use"])
        tm.epoch = format_second(self.epochtime["use"])
        tm.avg = format_second(avg)
        tm.last = format_second(last)
        trainer.logger.info(tm)

        super().on_train_epoch_end(trainer, func, params, meter, *args, **kwargs)

    def on_train_batch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        else:
            if self.avg:
                self.meter.update(meter)
                meter = self.meter
        trainer.logger.inline("{}/{}".format(params.idx + 1, len(trainer.train_dataloader)), meter, fix=1)

    def on_first_exception(self, trainer: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        trainer.logger.error("{} raised".format(e.__class__.__name__))

    def on_test_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        trainer.logger.info("tests start")

    def on_eval_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        trainer.logger.info("eval start")

    def on_eval_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        trainer.logger.info("eval end", meter)

    def on_test_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if meter is None:
            meter = ""
        trainer.logger.info("tests end", meter)


class ModelCheckpoint(TrainCallback):
    """
    用于检视训练过程中模型的某个指标，并根据其提升进行 checkpoint 类型的保存
    该类参考了 Keras 中相应的实现。
    """
    only_main_process = True

    def __init__(self, monitor, mode="train", lower=True, start_epoch=0):
        self.monitor = monitor
        self.mode = mode
        self.lower = lower
        self.last_val = NoneItem()
        self.start_epoch = start_epoch

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.update("train", trainer, params, meter)

    def update(self, cur_mode, trainer, param, meter):
        if cur_mode != self.mode:
            return
        if param.eidx > self.start_epoch:
            item = meter[self.monitor]
            if isinstance(item, AvgItem):
                item = item.avg
            if isinstance(item, NoneItem):
                return

            if self.lower:
                if self.last_val > item:
                    trainer.logger.info("model imporved from {} to {}".format(self.last_val, item))
                    trainer.save_checkpoint(meter.serialize())
                    self.last_val = item
            else:
                if self.last_val < item:
                    trainer.logger.info("model imporved from {} to {}".format(self.last_val, item))
                    trainer.save_checkpoint(meter.serialize())
                    self.last_val = item

    def on_test_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.update("test", trainer, params, meter)

    def on_eval_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.update("eval", trainer, params, meter)

    def __repr__(self) -> str:
        return self._repr_by_val("monitor", "mode", "lower", "start_epoch")


class TimingCheckpoint(TrainCallback):
    """
    在 Trainer 训练过程中定时保存模型
    """
    only_main_process = True

    def __init__(self, per_epoch=50):
        self.per_epoch = per_epoch

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if params.eidx % self.per_epoch == 0 and params.eidx > 0:
            trainer.save_keypoint(meter.serialize(), replacement=True)

    def __repr__(self) -> str:
        return self._repr_by_val("per_epoch")


class KeyErrorSave(TrainCallback):
    only_main_process = True
    only_single_gpu = True
    priority = -1

    def __init__(self, wait_input=False):
        self.wait_input = wait_input

    def on_first_exception(self, trainer: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        if isinstance(e, (KeyboardInterrupt)):
            trainer.logger.info("KeyErrorSave trigged, save checkpoint")
            trainer.save_keypoint({"mode": "KeyboardInterrupt"})

            tp = "n"
            if self.wait_input:
                tp = input("continue train step? (y/other)")

            if tp.lower() == "y":
                return True


class CUDAErrorHold(TrainCallback):
    """
    当 CUDA out of memory 出现时，挂住程序。

    一般而言，程序运行过程中 CUDA 会分配到最大显存，程序的显存占用一般会比较稳定，
    但仍然存在一些例外情况，显存的占用会飘忽不定，此时可以使用该 Callback 解决这一问题。
    """
    only_single_gpu = True
    priority = -1

    def __init__(self, wait_input=False):
        self.wait_input = wait_input

    def on_first_exception(self, trainer: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        if isinstance(e, (RuntimeError)):
            if "cuda" in str(e).lower() and 'out of memory' in str(e).lower():
                trainer.logger.info("CUDA Error trigged, enter to continue or type any to exit")
                tp = input("enter to continue or any other type to exit?")

                if len(tp.strip()) == 0:
                    return True


class AutoRecord(TrainCallback):
    """
    自动记录训练过程中的所有变量到 tensorboard 中（epoch 级）
    """
    only_main_process = True
    priority = 100

    def __init__(self) -> None:
        super().__init__()
        from collections import defaultdict
        self._ignore_dict = defaultdict(set)

    def on_hooked(self, trainer: Trainer, params: Params):
        self.start = 0
        trainer.experiment.add_tag("record", 'writer')

    def ignore_key(self, mode, key):
        self._ignore_dict[mode].add(key)

    def _key_name(self, mode, key):
        return "{}_{}_".format(key, mode)

    def on_test_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if isinstance(meter, Meter):
            for k, v in meter.numeral_items():
                if k in self._ignore_dict[_ML.test]:
                    continue
                trainer.writer.add_scalar(self._key_name("test", k), v, params.eidx)

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.start = params.eidx

    def on_eval_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        if isinstance(meter, Meter):
            for k, v in meter.numeral_items():
                if k in self._ignore_dict[_ML.eval]:
                    continue
                trainer.writer.add_scalar(self._key_name("eval", k), v, params.eidx)

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: AvgMeter, *args, **kwargs):
        if isinstance(meter, Meter):
            for k, v in meter.numeral_items():
                if k in self._ignore_dict[_ML.train]:
                    continue
                trainer.writer.add_scalar(self._key_name("train", k), v, params.eidx)


class EMAUpdate(TrainCallback):
    only_main_process = True

    def on_train_batch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_train_batch_end(trainer, func, params, meter, *args, **kwargs)
        for k, v in trainer.model_dict.items():
            if k.lower().startswith('ema'):
                v.step()


class LRSchedule(TrainCallback):
    def __init__(self, schedule: Schedule = None, apply=True, use_eidx=True):
        self.schedule = schedule
        self.apply = apply
        self.use_eidx = use_eidx

    def on_hooked(self, trainer: Trainer, params: Params):
        super().on_hooked(trainer, params)
        if self.schedule is None:
            if 'lr_sche' not in params:
                trainer.logger.warn('lr_sche not exists in params and be assigned, {} will be unhooked after.')
                self.unhook()
            else:
                self.schedule = params.lr_sche

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_train_epoch_end(trainer, func, params, meter, *args, **kwargs)
        for k, v in trainer.optimizer_dict.items():
            if self.use_eidx:
                step = params.eidx
            else:
                step = params.global_step

            if self.apply:
                new_lr = self.schedule.apply(v, step)
                trainer.logger.info('{}.lr = {}'.format(k, new_lr))
            else:
                ratio = self.schedule.scale(v, step)
                trainer.logger.info('lr scale ratio = {}'.format(k, ratio))


class SuccessQuery(TrainCallback):
    """allow you exit by type KeyboardInterupt in success mode"""
    only_single_gpu = True

    def on_first_exception(self, trainer: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        super().on_first_exception(trainer, func, params, e, *args, **kwargs)
        if isinstance(e, (KeyboardInterrupt)):
            tp = input("success?(Y/N, default N)")
            if tp.lower() == 'y':
                trainer.experiment.end()
                trainer.stop_train()
                trainer.stop_current_epoch()
                return True
            else:
                return False


class ReportSche(TrainCallback):
    """
    log `schedule` in every epoch end
    `schedule` means `Schedule` in Params and have `sche` in the name, which will have different value in every epoch
    """
    only_main_process = True
    priority = 100

    def on_hooked(self, trainer: Trainer, params: Params):
        self.sche_lis = []
        for k, v in params.items():  # type:str, Any
            if isinstance(v, (Schedule, ScheduleList)) and 'sche' in k.lower():
                self.sche_lis.append((k, v))

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_train_epoch_end(trainer, func, params, meter, *args, **kwargs)
        m = Meter()
        for k, v in self.sche_lis:
            m[k] = v(params.eidx)
        trainer.logger.info(m)



"""
for reading values recorded by tensorboard directly instead of opening tensorboard web pages.
"""
import pprint as pp
from collections import namedtuple
from typing import List
from tensorboard.backend.event_processing import event_accumulator

Scalars = namedtuple('ScalarEvent', ['wall_times', 'values', 'steps'])


class BoardReader():
    """
    ScalarEvent = namedtuple('ScalarEvent', ['wall_time', 'step', 'value'])
    CompressedHistogramEvent = namedtuple('CompressedHistogramEvent',
                                          ['wall_time', 'step',
                                           'compressed_histogram_values'])
    HistogramEvent = namedtuple('HistogramEvent',
                                ['wall_time', 'step', 'histogram_value'])
    HistogramValue = namedtuple('HistogramValue', ['min', 'max', 'num', 'sum',
                                                   'sum_squares', 'bucket_limit',
                                                   'bucket'])
    ImageEvent = namedtuple('ImageEvent', ['wall_time', 'step',
                                           'encoded_image_string', 'width',
                                           'height'])
    AudioEvent = namedtuple('AudioEvent', ['wall_time', 'step',
                                           'encoded_audio_string', 'content_type',
                                           'sample_rate', 'length_frames'])
    TensorEvent = namedtuple('TensorEvent', ['wall_time', 'step', 'tensor_proto'])
    """

    def __init__(self, file):
        self.ea = event_accumulator.EventAccumulator(file,
                                                     size_guidance={  # see below regarding this argument
                                                         event_accumulator.COMPRESSED_HISTOGRAMS: 500,
                                                         event_accumulator.IMAGES: 4,
                                                         event_accumulator.AUDIO: 4,
                                                         event_accumulator.SCALARS: 0,
                                                         event_accumulator.HISTOGRAMS: 1,
                                                     })
        self._reloaded = False

    def _check_reload(self):
        """load values from dist to memory"""
        if not self._reloaded:
            self.ea.Reload()

    def get_scalars(self, tag) -> Scalars:
        """get scalars named '<tag>' in this board"""
        self._check_reload()
        wall_times, steps, values = list(zip(*self.ea.Scalars(tag)))  # 'wall_time', 'step', 'value'
        values = [float("{:.4f}".format(i)) for i in values]
        return Scalars(wall_times, values, steps)

    @property
    def scalars_tags(self) -> List[str]:
        """get all scalar tags"""
        self._check_reload()
        return self.tags['scalars']

    @property
    def tags(self) -> dict:
        """get all tags in this board"""
        self._check_reload()
        return self.ea.Tags()

    def summary(self):
        """print tags in this board"""
        self._check_reload()
        pp.pprint(self.ea.Tags())



"""
runtime environment constant.

Pycharm debug mode currently has a bug that can't run next step after a breakpoint when have multithread(num_workers>0).

So the parameter can be conditional set as:

```python
from thexp import ENV
num_workers = 0 is ENV.IS_PYCHARM_DEBUG else 4
```

The class ENVIRON_ is also embeded as a class variable named `ENV` in class Params, means you can use like:

```python
from thexp import Params
Params.ENV.IS_PYCHARM_DEBUG
```
"""
import os
from sys import platform


class ENVIRON_:
    IS_WIN = (platform == "win32")
    IS_MAC = (platform == "darwin")
    IS_LINUX = (platform == "linux" or platform == "linux2")

    IS_REMOTE = any([i in os.environ for i in ['SHELL',
                                               'SHLVL',
                                               'SSH_CLIENT',
                                               'SSH_CONNECTION',
                                               'SSH_TTY']])
    IS_LOCAL = not IS_REMOTE

    IS_PYCHARM = os.environ.get("PYCHARM_HOSTED", 0) == "1"
    IS_PYCHARM_DEBUG = eval(os.environ.get('IPYTHONENABLE', "False"))





"""

"""

from thexp.frame.saver import Saver

saver = Saver("./sav")

def test_max_to_keep():
    for i in range(5):
        saver.save_checkpoint(i,dict(a=i),dict(b=i))
    assert len(saver.find_checkpoints()) == saver.max_to_keep
    saver.clear_checkpoints()
    assert len(saver.find_checkpoints()) == 0

# def test_keypoint():
#     saver.clear_keypoints()
#     for i in range(5):
#         saver.save_keypoint(i,dict(a=i),dict(b=i))
#     for i in range(5):
#         assert saver.load["a"] == i
#         assert saver.load_keypoint_info(i)["b"] == i
#
#     assert len(saver.find_keypoints()) == 5
#     saver.clear_keypoints()
#     assert len(saver.find_keypoints()) == 0



# TODO colored text
import ctypes
import sys

if sys.platform == "win32" and sys.stdin.isatty():
    FOREGROUND_BLACK = 0x00  # black.
    FOREGROUND_DARKBLUE = 0x01  # dark blue.
    FOREGROUND_DARKGREEN = 0x02  # dark green.
    FOREGROUND_DARKSKYBLUE = 0x03  # dark skyblue.
    FOREGROUND_DARKRED = 0x04  # dark red.
    FOREGROUND_DARKPINK = 0x05  # dark pink.
    FOREGROUND_DARKYELLOW = 0x06  # dark yellow.
    FOREGROUND_DARKWHITE = 0x07  # dark white.
    FOREGROUND_DARKGRAY = 0x08  # dark gray.
    FOREGROUND_BLUE = 0x09  # blue.
    FOREGROUND_GREEN = 0x0a  # green.
    FOREGROUND_SKYBLUE = 0x0b  # skyblue.
    FOREGROUND_RED = 0x0c  # red.
    FOREGROUND_PINK = 0x0d  # pink.
    FOREGROUND_YELLOW = 0x0e  # yellow.
    FOREGROUND_WHITE = 0x0f  # white.
    FOREGROUND_PURPLE = FOREGROUND_PINK
    FOREGROUND_CYAN = FOREGROUND_PINK

    BACKGROUND_BLACK = 0x00
    BACKGROUND_DARKBLUE = 0x10  # dark blue.
    BACKGROUND_DARKGREEN = 0x20  # dark green.
    BACKGROUND_DARKSKYBLUE = 0x30  # dark skyblue.
    BACKGROUND_DARKRED = 0x40  # dark red.
    BACKGROUND_DARKPINK = 0x50  # dark pink.
    BACKGROUND_DARKYELLOW = 0x60  # dark yellow.
    BACKGROUND_DARKWHITE = 0x70  # dark white.
    BACKGROUND_DARKGRAY = 0x80  # dark gray.
    BACKGROUND_BLUE = 0x90  # blue.
    BACKGROUND_GREEN = 0xa0  # green.
    BACKGROUND_SKYBLUE = 0xb0  # skyblue.
    BACKGROUND_RED = 0xc0  # red.
    BACKGROUND_PINK = 0xd0  # pink.
    BACKGROUND_YELLOW = 0xe0  # yellow.
    BACKGROUND_WHITE = 0xf0  # white.
    BACKGROUND_PURPLE = BACKGROUND_PINK
    BACKGROUND_CYAN = BACKGROUND_PINK
else:
    FOREGROUND_BLACK = 30
    FOREGROUND_RED = 31
    FOREGROUND_GREEN = 32
    FOREGROUND_YELLOW = 33
    FOREGROUND_BLUE = 34
    FOREGROUND_PURPLE = 35
    FOREGROUND_CYAN = 36
    FOREGROUND_WHITE = 37
    FOREGROUND_SKYBLUE = FOREGROUND_BLUE
    FOREGROUND_PINK = FOREGROUND_PURPLE
    FOREGROUND_DARKBLUE = FOREGROUND_BLUE
    FOREGROUND_DARKGREEN = FOREGROUND_GREEN
    FOREGROUND_DARKSKYBLUE = FOREGROUND_BLUE
    FOREGROUND_DARKRED = FOREGROUND_RED
    FOREGROUND_DARKPINK = FOREGROUND_CYAN
    FOREGROUND_DARKYELLOW = FOREGROUND_YELLOW
    FOREGROUND_DARKWHITE = FOREGROUND_WHITE
    FOREGROUND_DARKGRAY = FOREGROUND_BLACK

    BACKGROUND_BLACK = 40
    BACKGROUND_RED = 41
    BACKGROUND_GREEN = 42
    BACKGROUND_YELLOW = 43
    BACKGROUND_BLUE = 44
    BACKGROUND_PURPLE = 45
    BACKGROUND_CYAN = 46
    BACKGROUND_WHITE = 47
    BACKGROUND_SKYBLUE = BACKGROUND_BLUE
    BACKGROUND_PINK = BACKGROUND_PURPLE
    BACKGROUND_DARKBLUE = BACKGROUND_BLUE
    BACKGROUND_DARKGREEN = BACKGROUND_GREEN
    BACKGROUND_DARKSKYBLUE = BACKGROUND_BLUE
    BACKGROUND_DARKRED = BACKGROUND_RED
    BACKGROUND_DARKPINK = BACKGROUND_PURPLE
    BACKGROUND_DARKYELLOW = BACKGROUND_YELLOW
    BACKGROUND_DARKWHITE = BACKGROUND_WHITE
    BACKGROUND_DARKGRAY = BACKGROUND_BLACK

MODE_NORMAL = 0
MODE_BOLD = 1
MODE_UNDERLINE = 4
MODE_BLINK = 5
MODE_INVERT = 7
MODE_HIDE = 8

if sys.platform == "win32" and sys.stdin.isatty():
    STD_INPUT_HANDLE = -10
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12
    std_out_handle = ctypes.windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)
    std_err_handle = ctypes.windll.kernel32.GetStdHandle(STD_ERROR_HANDLE)
    # _handle_dict = dict(
    #     stdout=std_out_handle,
    #     stderr=std_err_handle
    # )
else:
    std_out_handle = sys.stdout
    std_err_handle = sys.stderr
_handle_dict = dict(
    stdout=sys.stdout,
    stderr=sys.stderr
)


def set_cmd_text_color(color, handle=std_out_handle):
    Bool = ctypes.windll.kernel32.SetConsoleTextAttribute(handle, color)
    return Bool


def reset_color():
    set_cmd_text_color(FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE)


def in_terminal():
    '''
    在win环境中用python xx.py命令时，只支持获取句柄更改的手段，其他时候，可以用 \033
    :return:
    '''
    return sys.stdin.isatty()


def in_win():
    return sys.platform == "win32"


def cprint(*args, fore=None, back=None, sep=" ", end="", handle="stdout", flush=True):
    value = sep.join(*args)

    if fore is None and back is None:
        reset_color()
    elif back is None:
        set_cmd_text_color(fore)
    elif fore is None:
        set_cmd_text_color(back)
    else:
        set_cmd_text_color(fore | back)

    print(value, end=end, file=_handle_dict[handle], flush=flush)
    # sys.stdout.write(str(value))
    # if flush:
    #     sys.stdout.flush()

    reset_color()
    return cprint


# def eprint(*args, fore=None, back=None, mode=None, sep=" ", end="", handle="stdout", flush=True):
#     value = sep.join(*args)
#
#     if mode is None:
#         mode = MODE_NORMAL
#     if fore is None and back is None:
#         print(f"\033[{mode}m{value}\033[0m", end=end, file=_handle_dict[handle], flush=flush)
#     elif back is None:
#         print(f"\033[{mode};{fore}m{value}\033[0m", end=end, file=_handle_dict[handle], flush=flush)
#     elif fore is None:
#         print(f"\033[{mode};;{back}m{value}\033[0m", end=end, file=_handle_dict[handle], flush=flush)
#     else:
#         print(f"\033[{mode};{fore};{back}m{value}\033[0m", end=end, file=_handle_dict[handle], flush=flush)


def uprint(*args,
           fore=None,
           back=None,
           mode=None,
           sep=" ",
           end="",
           handle="stdout",
           flush=True):
    '''
    Prints the colored values to sys.stdout or sys.stderr.
    :param fore:
    :param back:
    :param mode:
    :param sep: string inserted between values, default a space.
    :param end: string appended after the last value, default a newline.
    :param handle: str, "stdout" or "stderr"
        note:the stderr haven't be tested.
    :param flush: whether to forcibly flush the stream.
    :param continuous: default True, if True, the return value will be a lambda method wrapped the uprint mathod
                and the options.
    :return: uprint method, you can use
        uprint()()()() to output strings with diff color.
    '''
    args = [str(i) for i in args]
    if in_win() and in_terminal():
        cprint(args, fore=fore, back=back, sep=sep, end=end, handle=handle, flush=flush)
    else:
        pass
        # eprint(args, fore=fore, back=back, mode=mode, sep=sep, end=end, handle=handle, flush=flush)

    def extendprint(*args,
                    fore=fore,
                    back=back,
                    mode=mode,
                    sep=sep,
                    end=end,
                    handle=handle,
                    flush=flush):
        return uprint(*args, fore=fore, back=back, mode=mode, sep=sep, end=end, handle=handle, flush=flush)

    return extendprint



from setuptools import setup, find_packages
from thexp import __VERSION__
from rename_templete import copy_templete

copy_templete()

setup(
    name='thexp',
    # 主版本，次版本，修订号？，bug修订号，...待定
    version=__VERSION__,
    description='An useful torch framework for dl experiments.',
    url='https://github.com/sailist/thexp',
    author='sailist',
    author_email='sailist@outlook.com',
    license='Apache License 2.0',
    include_package_data=True,
    install_requires=[
        'torch',
        'tensorboard', 'matplotlib', 'numpy', 'pandas', 'scipy', 'sklearn',
        'fire',
        'gitpython',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='thexp',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'thexp = thexp.cli.cli:main'
        ]
    },
)



"""

"""

from .builder import DatasetBuilder, Delegate
from .databundler import DataBundler
from .experiment import Experiment
from .configs import globs
from .logger import Logger
from .meter import Meter, AvgMeter
from .params import Params, BaseParams
from .rndmanager import RndManager
from .saver import Saver
from .trainer import Trainer, BaseTrainer



"""

"""
import os

from ..base_classes.defaults import draw_dict
from ..utils.dates import curent_date
from ..decorators import deprecated


@deprecated('1.4.0', '1.5.0')
class Reporter():
    """
    记录并生成 markdown 格式的报告
    """

    def __init__(self, pltf_dir, exts=None):

        self.base_dir = pltf_dir
        from collections import defaultdict
        self.cur_key = curent_date()
        self.plot_dict = defaultdict(draw_dict)
        self.cur_dir = None
        self.ftags = []
        self.tags = set()

    @property
    def savedir(self):
        if self.cur_dir is None:
            i = 1
            fn = os.path.join(self.base_dir, '{:04}'.format(i))
            while os.path.exists(fn):
                i += 1
                fn = os.path.join(self.base_dir, '{:04}'.format(i))
            os.makedirs(fn, exist_ok=True)
            self.cur_dir = fn

        return self.cur_dir

    def as_numpy(self, var):
        import torch, numpy as np
        if isinstance(var, (int, float)):
            return var

        if isinstance(var, torch.Tensor):
            return var.detach().cpu().item()

        if isinstance(var, np.ndarray):
            return var.item()

        assert False

    def add_scalar(self, var, step, tag):
        self.plot_dict[tag]['x'].append(step)
        self.plot_dict[tag]['y'].append(self.as_numpy(var))

    def add_ftag(self, file, tag):

        self.ftags.append((os.path.normpath(file), tag))

    def savefig(self):
        from matplotlib import pyplot as plt
        from thexp.utils.paths import filter_filename
        dir = self.savedir
        for k, v in self.plot_dict.items():
            if 'fn' in v:
                continue

            plt.figure()
            base_name = filter_filename('{}.jpg'.format(k))
            fn = os.path.join(dir, base_name)
            plt.plot(v['x'], v['y'])
            plt.title(k)
            plt.savefig(fn)
            v['fn'] = base_name
        return dir

    @property
    def picklefile(self):
        return 'analyser.pkl'

    @property
    def reportfile(self):
        return 'report.md'

    def add_tag(self, tag):
        self.tags.add(str(tag))

    def report(self, every=20, otherinfo=None):
        every = max(every, 1)

        from thexp.utils.markdown import Markdown
        md = Markdown()
        md.add_title(curent_date('%y-%m-%d-%H:%M:%S'))
        with md.code() as code:
            for tag in self.tags:
                code.add_line(tag)

        if otherinfo is not None:
            md.extends(otherinfo)
        md.add_title('Vars', level=2)
        self.savefig()
        for k, v in self.plot_dict.items():
            md.add_title(k, level=3)
            if 'fn' in v:
                md.add_picture('./{}'.format(v['fn']), k, False)

            lines = []
            head = ['step']
            vars = ['values']

            xs = []
            ys = []

            for x, y in zip(v['x'][::every], v['y'][::every]):
                head.append('**{}**'.format(x))
                vars.append('{:.4f}'.format(y))
                xs.append(x)
                ys.append(y)
                if len(head) > 10:
                    lines.append(head)
                    lines.append(vars)
                    head = ['**step**']
                    vars = ["values"]

            if len(head) != 1:
                lines.append(head)
                lines.append(vars)

            ys = [float("{:.4f}".format(y)) for y in v["y"]]
            md.add_table(lines)
            code = """
            every = {}
            xs = {}
            ys = {}
            plt.plot(xs[::every],ys[::every])
            plt.title({})
           """.format(every, v["x"], ys, k)
            md.add_code(code, "python")

        md.add_title("Files", level=2)
        with md.quote() as q:
            q.add_text("PWD : {}".format(os.getcwd()))
        with md.table() as table:
            table.head(["index", "tag", "file"])
            for i, (file, tag) in enumerate(self.ftags):
                table.append([i, tag, file])

        dir = self.savedir
        fn = os.path.join(dir, self.reportfile)
        md.to_file(fn)
        return fn



"""

"""
import bisect
import os
import pprint as pp
import warnings

# from ..utils.lazy import torch, np
import numpy as np
import torch
from functools import lru_cache
from functools import wraps
from torch.utils.data.dataloader import DataLoader
from torch.optim.optimizer import Optimizer
from typing import Any, Union, List, Dict
from collections.abc import Iterator
from .databundler import DataBundler
from .meter import AvgMeter, Meter
from .params import Params
from .saver import Saver
from ..base_classes.metaclasses import Merge
from ..globals import _BUILTIN_PLUGIN, _FNAME, _PLUGIN_DIRNAME, _PLUGIN_KEY, _OS_ENV


def mp_agent(rank, self, op):
    import torch.distributed as dist
    self.params.local_rank = rank
    dist.init_process_group(backend='nccl', init_method=self.params.init_method,
                            rank=rank,
                            world_size=self.params.world_size)

    self.params.device = 'cuda:{}'.format(rank)
    self.regist_device(torch.device(self.params.device))
    torch.cuda.set_device(self.params.local_rank)
    print('in rank {}'.format(rank))
    self.models(self.params)
    self.datasets(self.params)
    self.callbacks(self.params)
    op(self)


class BaseTrainer(metaclass=Merge):
    __exp_name__ = None
    _call_backs = {
        "initial",
        "train", "train_epoch", "train_step", "test", "eval", "train_on_batch",
        "regist_databundler", "train_batch", "test_eval_logic", "test_eval_logic_v2", "predict",
        "load_keypoint", "load_checkpoint", "load_model", "save_keypoint", "save_checkpoint", "save_model",
    }

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        if cls.__exp_name__ is None:
            cls.__exp_name__ = cls.__name__.lower().replace("trainer", "Exp")

        def wrapper(func, _call_set: list):
            """
            对每个 Trainer 的 _call_backs 类变量中定义的函数尝试绑定回调
            Args:
                func:
                _call_set:

            Returns:

            """

            @wraps(func)
            def _newfunc(*aargs, **kkwargs):
                """执行前回调 on_begin() 、执行后回调 on_end()、执行异常则回调 on_exception() """
                for callback in _call_set:
                    callback.on_begin(self, func, self.params, *aargs, **kkwargs)
                try:
                    _meter = func(*aargs, **kkwargs)
                    # 如果返回迭代器，那么会消耗掉迭代器，并只返回最后一次运行的结果（如果有）
                    if isinstance(_meter, Iterator):
                        _m = Meter()
                        for _m in _meter:
                            pass
                        _meter = _m
                except BaseException as e:
                    _handles = [callback.on_exception(self, func, self.params, e, *aargs, **kkwargs)
                                for callback in _call_set]

                    if any(_handles):
                        return None
                    else:
                        raise e

                for callback in _call_set:
                    callback.on_end(self, func, self.params, _meter, *aargs, **kkwargs)
                return _meter

            return _newfunc

        self._callback_set = []
        self._callback_name_set = set()

        vars = dir(self)
        for name in vars:
            if name not in self._call_backs:
                continue
            if name.startswith("_"):
                continue
            value = getattr(self, name, None)
            if value is None:
                continue
            if callable(value):
                setattr(self, name, wrapper(value, self._callback_set))
        return self

    def __init__(self, params: Params = None):
        self._model_dict = {}  # type:Dict[str,torch.nn.Module]
        self._optim_dict = {}  # type:Dict[str,Optimizer]
        self._other_state_dict = {}
        self._vector_dict = {}
        self._checkpoint_plug = {}
        self._databundler_dict = {}  # type:Dict[str,DataBundler]
        self.train_epoch_toggle = False
        self.train_toggle = False
        self.experiment = None
        if params is not None:
            self.params = params
            if isinstance(params.device, str):
                _device = torch.device(params.device)
                self.regist_device(_device)
                if 'cuda' in params.device:
                    torch.cuda.set_device(_device)
            else:
                assert False
            # elif isinstance(params.device, (list, dict)):
            #     if isinstance(params.device, list):
            #         self.regist_devices([torch.device(i) for i in params.device])
            #     elif isinstance(params.device, dict):
            #         self.regist_devices({k: torch.device(v) for k, v in params.device.items()})
            #     elif isinstance(params.device, torch.device):
            #         warnings.warn("define torch.device in params is not recommanded, allocate a string is better.")
            #         self.regist_device(params.device)
            #     else:
            #         warnings.warn("Unknown type for params.device.")

            if params.contains('tmp_dir'):
                if params.tmp_dir is not None:
                    os.environ['TMPDIR'] = params.tmp_dir

            if params.local_rank >= 1:
                from thexp import globs
                globs['rank'] = params.local_rank
        else:
            self.params = Params()
        self.initial()

    def __setstate__(self, state):
        self._model_dict = state['_model_dict']
        self._optim_dict = state['_optim_dict']
        self._other_state_dict = state['_other_state_dict']
        self._vector_dict = state['_vector_dict']
        self._checkpoint_plug = state['_checkpoint_plug']
        self._databundler_dict = state['_databundler_dict']
        self.train_epoch_toggle = state['train_epoch_toggle']
        self.train_toggle = state['train_toggle']
        self.params = state['params']
        self.experiment = state.get('experiment', None)

    def __getstate__(self):
        res = {
            '_model_dict': self._model_dict,
            '_optim_dict': self._optim_dict,
            '_other_state_dict': self._other_state_dict,
            '_vector_dict': self._vector_dict,
            '_checkpoint_plug': self._checkpoint_plug,
            '_databundler_dict': self._databundler_dict,
            'train_epoch_toggle': self.train_epoch_toggle,
            'train_toggle': self.train_toggle,
            'experiment': self.experiment,
            'params': self.params,
        }
        empk = [k for k in res if res[k] is None]
        for k in empk:
            res.pop(k)
        return res

    def initial(self):
        """initial the trainer"""
        import inspect
        from .experiment import Experiment
        # build experiment
        self.params.initial()
        file = inspect.getfile(self.__class__)
        dirname = os.path.basename(os.path.dirname(file))

        pre = os.path.splitext(os.path.basename(file))[0]

        if not self.params.get('git_commit', True):
            os.environ[_OS_ENV.THEXP_COMMIT_DISABLE] = '1'

        self.experiment = Experiment("{}.{}".format(pre, dirname))

        # rigist and save params of this training procedure
        self.experiment.add_params(self.params)

        # regist trainer info
        trainer_kwargs = {
            _PLUGIN_KEY.TRAINER.path: inspect.getfile(self.__class__),
            _PLUGIN_KEY.TRAINER.doc: self.__class__.__doc__,
            _PLUGIN_KEY.TRAINER.fn: pre,
            _PLUGIN_KEY.TRAINER.class_name: self.__class__.__name__
        }
        self.experiment.add_plugin(_BUILTIN_PLUGIN.trainer, trainer_kwargs)

        self.callbacks(self.params)
        self.models(self.params)
        self.datasets(self.params)

    def _regist_databundler(self, key, val):
        from torch.utils.data import DataLoader
        assert isinstance(val, (DataBundler, DataLoader))
        if isinstance(val, DataLoader):
            val = DataBundler().add(val)

        # To ensure that children threads(dataloader workers)  will be killed
        if key in self._databundler_dict:
            del self._databundler_dict[key]

        self._databundler_dict[key] = val

    def regist_device(self, device: torch.device):
        self.device = device

    def regist_databundler(self,
                           train: Union[DataBundler, DataLoader] = None,
                           eval: Union[DataBundler, DataLoader] = None,
                           test: Union[DataBundler, DataLoader] = None):
        """
        regist train/eval/test dataloader

        Args:
            train / eval / test: DataBundler in thexp, or DataLoader in pytorch
                if None, the corresponding train/eval/test methods will be ignored when call them.
        """
        if train is not None:
            self._regist_databundler("train", train)
        if eval is not None:
            self._regist_databundler("eval", eval)
        if test is not None:
            self._regist_databundler("tests", test)

        self.logger.info(self._databundler_dict)

    def stop_train(self):
        """stop current training procedure"""
        self.train_toggle = True

    def stop_current_epoch(self):
        """stop current training epoch, if current epoch doesn't reach the end,
        the next epoch will be started and the releated callback methods will be called.
        """
        self.train_epoch_toggle = True

    def train(self):
        params = self.params
        while params.eidx < params.epoch + 1:
            self.train_epoch(params.eidx, params)
            params.eidx += 1
            if self.train_toggle:
                self.train_toggle = False
                break

    def train_epoch(self, eidx: int, params: Params):
        avg = AvgMeter()
        self.change_mode(True)
        for idx, batch_data in enumerate(self.train_dataloader):  # 复现多线程下 Keyboard Interupt，尝试通过Try解决
            meter = self.train_batch(eidx, idx, self.params.global_step, batch_data, params, self.device)
            avg.update(meter)
            # del meter

            params.global_step += 1
            params.idx = idx
            if self.train_epoch_toggle:
                self.train_epoch_toggle = False
                break

        self.change_mode(False)
        return avg

    def train_step(self, steps) -> Union[AvgMeter, Meter]:
        """
        train specific steps
        Args:
            steps: int

        Returns:
            if steps > 1, will return a AveMeter instance.
            if steps = 1, will return a Meter object.

        """
        param = self.params
        i = 0
        if steps == 1:
            for idx, data in enumerate(self.train_dataloader):
                return self.train_batch(0, idx, i, data, param, self.device)

        avg = AvgMeter()
        while steps > 0:
            avg = AvgMeter()
            for idx, data in enumerate(self.train_dataloader):
                meter = self.train_batch(0, idx, i, data, param, self.device)
                steps -= 1
                avg.update(meter)
                if steps <= 0:
                    return avg
        return avg

    def feed_batchdata(self, batch_data=None) -> Meter:
        """
        train a step for testing or some other purpose.
        Args:
            batch_data: the batch_data used in training procedure.
                if None, batch_data will be fetched from train_datasetloader.

        Returns:
            a Meter.
        """
        if batch_data is None:
            return self.train_step(1)
        return self.train_batch(0, 0, 0, batch_data, self.params, self.device)

    def test(self):
        """test via test_dataloader"""

        loader = self.test_dataloader
        if loader is None:
            self.logger.info("Have no test dataset, ignored test.")
            return None
        return self.test_eval_logic_v2(loader, self.params, True)

    def eval(self):
        """eval via eval_dataloader"""
        loader = self.eval_dataloader
        if loader is None:
            self.logger.info("Have no eval dataset, ignored eval.")
            return None
        return self.test_eval_logic_v2(loader, self.params, False)

    @property
    def in_main_process(self):
        return self.params.local_rank <= 0

    @property
    def safe_writer(self):
        """see trainer.writer"""
        import tensorflow as tf
        import tensorboard as tb
        tf.io.gfile = tb.compat.tensorflow_stub.io.gfile
        return self.writer

    @property
    @lru_cache()
    def writer(self):
        """
        Notes:
        ------
        When using add_embedding, there may raise some exceptions cased by version conflict, here is some solutions:

        1. tensorflow_core._api.v1.io.gfile' or'tensorflow_core._api.v2.io.gfile' has no attribute 'get_filesystem'
        first, try upgrade tensorboard and tensorflow as followed version:
            tensorboard==2.0.2
            tensorflow==2.0.0

        if you still have the same problem, use this code as a temporary solution:

            import tensorflow as tf
            import tensorboard as tb
            tf.io.gfile = tb.compat.tensorflow_stub.io.gfile

        use `trainer.safe_writer` to get a writter with these code added inside thexp.

        solution is referred by https://github.com/pytorch/pytorch/issues/30966


        2. You may cause PermissionError like: [Errno 13] Permission denied: '/tmp/.tensorboard-info/pid-20281.info'
        the solution is to set environment variable TMPDIR

            export TMPDIR=/tmp/$USER;
            mkdir -p $TMPDIR;
            tensorboard --logdir ...

        code in line:
            export TMPDIR=/tmp/$USER; mkdir -p $TMPDIR; tensorboard --logdir ....

        solution is referred by https://github.com/tensorflow/tensorboard/issues/2010

        Returns:
            A SummaryWriter instance
        """
        from torch.utils.tensorboard import SummaryWriter

        kwargs = self.experiment.add_board()
        res = SummaryWriter(**kwargs)

        def close(*args):
            res.flush()
            res.close()

        self.experiment.regist_exit_hook(close)
        return res

    @property
    @lru_cache()
    def logger(self):
        """see thexp.frame.Logger"""
        from .logger import Logger
        logger = Logger()
        fn = logger.add_log_dir(self.experiment.test_dir)
        self.experiment.add_logger(fn)
        return logger

    @property
    @lru_cache()
    def saver(self):
        """see thexp.frame.Saver"""
        kwargs = self.experiment.add_saver()
        return Saver(**kwargs)

    @property
    @lru_cache()
    def rnd(self):
        """see thexp.frame.RndManager"""
        from .rndmanager import RndManager
        kwargs = self.experiment.add_rndmanager()
        return RndManager(**kwargs)

    @property
    def model_dict(self) -> Dict[str, torch.nn.Module]:
        return self._model_dict

    @property
    def optimizer_dict(self) -> Dict[str, Optimizer]:
        return self._optim_dict

    @property
    def train_dataloader(self) -> DataBundler:
        return self._databundler_dict.get("train", None)

    @property
    def eval_dataloader(self) -> DataBundler:
        return self._databundler_dict.get("eval", None)

    @property
    def test_dataloader(self) -> DataBundler:
        return self._databundler_dict.get("tests", None)

    @classmethod
    def from_params(cls, params: Params = None):
        return cls(params)

    def regist_checkpoint(self, key, func):
        """
        注册需要被 checkpoint 类型的字典保存的
        Args:
            key:
            func:

        Returns:

        """
        self._checkpoint_plug[key] = func

    def save_keypoint(self, extra_info=None, replacement=False):
        """
        保存 keypoint，会保存所有可存储格式
        Args:
            extra_info:  额外的信息，将以 json 格式被保存在和模型文件名相同，但后缀名为 json 的文件中
            replacement: 若遇到相同文件名，是否进行替换

        Returns:

        """
        state_dict = self.checkpoint_dict()
        fn = self.saver.save_keypoint(self.params.eidx, state_dict, extra_info, replacement)
        self.logger.info("save keypoint in {}".format(fn))
        return fn

    def save_checkpoint(self, extra_info=None, replacement=False):
        """
        保存 checkpoint，会保存所有可存储格式
        Args:
            extra_info:  额外的信息，将以 json 格式被保存在和模型文件名相同，但后缀名为 json 的文件中
            replacement: 若遇到相同文件名，是否进行替换

        Returns:
            保存的 checkpoint 的文件名
        """
        state_dict = self.checkpoint_dict()
        fn = self.saver.save_checkpoint(self.params.eidx, state_dict, extra_info, replacement)
        self.logger.info("save checkpoint in {}".format(fn))
        return fn

    def save_model(self, extra_info: dict = None) -> str:
        """
        保存 model，只会保存所有 torch.nn.Module 类型的 state_dict
        Args:
            extra_info: 额外的信息，将以 json 格式被保存在和模型文件名相同，但后缀名为 json 的文件中

        Returns:
            所保存模型的文件名

        """
        state_dict = self.model_state_dict()
        fn = self.saver.save_model(self.params.eidx, state_dict, extra_info)
        self.logger.info("save model in {}".format(fn))
        return fn

    def add_callback(self, callback):
        """
        添加一个回调函数，注意，不能添加重复的 callback，这不推荐，也没有必要。
        :type callable,str
        :param callback:
        :return:
        """
        msg = None
        cb_name = callback.__class__.__name__
        if callback not in self._callback_set and cb_name in self._callback_name_set:
            msg = "Callback duplicate."
            callback.on_hook_failed(self, msg)

        if msg is not None:
            return False
        bisect.insort(self._callback_set, callback)
        self._callback_name_set.add(cb_name)

        callback._trainer = self
        callback.on_hooked(self, self.params)
        self.logger.info("{} hooked on {}.".format(callback, self))
        return True

    def reload_callback(self, callback):
        """重新加载某 callback"""
        self.remove_callback(callback.__class__)
        return self.add_callback(callback)

    def remove_callback(self, callback):
        """
        移除已加载的 callback
        Args:
            callback: 可以是回调类名、实例、或回调类类型

        Returns:
            是否移除成功，若返回False，则表明没有找到对应的 callback
            若返回 True，则表明该 callback 已被完好移除
        """
        msg = None
        from .callbacks import BaseCallback

        try:
            if issubclass(callback, BaseCallback):
                for cb in self._callback_set:
                    if cb.__class__.__name__ == callback.__name__:
                        callback = cb
                        break
        except:  # handle TypeError: issubclass() arg 1 must be a class
            pass

        if isinstance(callback, str):
            for cb in self._callback_set:
                if cb.__class__.__name__ == callback:
                    callback = cb
                    break

        if callback not in self._callback_set:
            return False

        cb_name = callback.__class__.__name__
        self._callback_set.remove(callback)
        self._callback_name_set.remove(cb_name)
        self.logger.info("{} unhooked from {}.".format(callback, self))
        return True

    '''module和optim的一部分方法集成'''

    def load_checkpoint(self, fn):
        ckpt, info = self.saver.load_state_dict(fn)
        self.load_checkpoint_dict(ckpt)
        self.logger.raw(pp.pformat(info))

    def load_model(self, fn, strict=True):
        ckpt, info = self.saver.load_state_dict(fn)
        self.load_model_state_dict(ckpt, strict=strict)
        self.logger.raw(pp.pformat(info))

    def load_checkpoint_dict(self, state_dict):
        self.logger.raw("loading checkpoint")
        self.params.eidx = state_dict['eidx']
        self.params.idx = state_dict['idx']
        self.params.global_step = state_dict['global_step']
        self.load_model_state_dict(state_dict["model"])
        self.load_optim_state_dict(state_dict["optim"])
        self.load_other_state_dict(state_dict["other"])
        self.load_vector_dict(state_dict["vector"])
        self.load_extra_state_dict(state_dict['plug'])

    def load_model_state_dict(self, state_dict, strict=True):
        self.logger.inline("loading model: ", append=True)
        for k in self._model_dict:
            self.logger.raw(k)
            if k in state_dict:
                self._model_dict[k].load_state_dict(state_dict[k], strict=strict)
            else:
                if strict:
                    raise KeyError(k)
                else:
                    warnings.warn("{} not found in state_dict".format(k))
        self.logger.newline()

    def load_optim_state_dict(self, state_dict, strict=False):
        self.logger.inline("loading optimizers: ", append=True)
        for k in self._optim_dict:
            self.logger.raw(k)
            if k in state_dict:
                self._optim_dict[k].load_state_dict(state_dict[k])
            else:
                if strict:
                    raise KeyError(k)
                else:
                    warnings.warn("{} not found in state_dict".format(k))
        self.logger.newline()

    def load_other_state_dict(self, state_dict, strict=False):
        self.logger.inline("loading other: ", append=True)
        for k in self._other_state_dict:
            self.logger.raw(k)
            if k in state_dict:
                self._other_state_dict[k].load_state_dict(state_dict[k])
            else:
                if strict:
                    raise KeyError(k)
                else:
                    warnings.warn("{} not found in state_dict".format(k))
        self.logger.newline()

    def load_vector_dict(self, state_dict, strict=False):
        self.logger.inline("loading vectors: ", append=True)
        for k in self._vector_dict:
            self.logger.raw(k)
            if k in state_dict:
                self.__setattr__(k, state_dict[k])
            else:
                if strict:
                    raise KeyError(k)
                else:
                    warnings.warn("{} not found in state_dict".format(k))
        self.logger.newline()

    def extra_state_dict(self) -> dict:
        """
        nn.Module, Optimizer, numpy.ndarray, torch.Tensor, 以及其他包含 state_dict 接口的对象在保存checkpoint时候
        无需手动添加，会自动被存储，而除了这些之外，有其他需要在 checkpoint 中被保存的内容，可以通过该接口实现
        Returns:

        Notes:
            该方法和 load_extra_state_dict() 一一对应，两者需要同时实现
        """
        return {}

    def load_extra_state_dict(self, state_dict, strict=False):
        """
        nn.Module, Optimizer, numpy.ndarray, torch.Tensor, 以及其他包含 state_dict 接口的对象在保存checkpoint时候
        无需手动添加，会自动被存储，而除了这些之外，有其他需要在 checkpoint 中被保存的内容，可以通过该接口实现
        Returns:

        Args:
            state_dict:
            strict:

        Returns:

        Notes:
            该方法和 extra_state_dict() 一一对应，两者需要同时实现
        """
        pass

    def load_state_dict(self, strict=True, **kwargs):
        """
        传入键值对，对 model / optim / other / checkpoint_plug 分别进行检查尝试，若能匹配则调用相应的加载方法
        Args:
            strict:  是否严格匹配，针对 model 和 checkpoint_plug ，当存在无法匹配的键时，
            若该值为 True，则抛出异常，否则仅报一次警告
            **kwargs:  键值对

        Returns:

        """
        for k, v in kwargs.items():
            if k in self._model_dict:
                self._model_dict[k].load_state_dict(v, strict)
            elif k in self._optim_dict:
                self._optim_dict[k].load_state_dict(v)
            elif k in self._other_state_dict:
                self._other_state_dict[k].load_state_dict(v)
            elif k in self._vector_dict:
                self.__setattr__(k, v)
            elif k in self._checkpoint_plug:
                self._checkpoint_plug[k](self, v, strict)
            elif strict:
                raise KeyError(k)
            else:
                warnings.warn("{} not found in all state_dict".format(k))

    def estimate_memory(self):
        for _, v in self._model_dict.items():
            pass

    def checkpoint_dict(self):
        val = dict(
            model=self.model_state_dict(),
            optim=self.optim_state_dict(),
            other=self.other_state_dict(),
            vector=self.vector_state_dict(),
            plug=self.extra_state_dict(),
            eidx=self.params.eidx,
            idx=self.params.idx,
            global_step=self.params.global_step,
            test_name=self.experiment.test_name,
        )
        return val

    def model_state_dict(self):
        """所有继承自 nn.module 的类的 state_dict """
        return {k: v.state_dict() for k, v in self._model_dict.items()}

    def optim_state_dict(self):
        """所有继承自 Optimizer 的类的 state_dict"""
        return {k: v.state_dict() for k, v in self._optim_dict.items()}

    def other_state_dict(self):
        """所有 实现了 state_dict / load_state_dict 接口的"""
        return {k: v.state_dict() for k, v in self._other_state_dict.items()}

    def vector_state_dict(self):
        """所有 torch.Tensor 或 numpy.ndarray"""
        return {k: v for k, v in self._vector_dict.items()}

    def change_mode(self, train=True):
        for k, v in self._model_dict.items():
            if train:
                v.train()
            else:
                v.eval()

    def to(self, device):
        for k, v in self._model_dict.items():
            self.__setattr__(k, v.to(device))
        for k, v in self._databundler_dict.items():
            v.to(device)
        for k, v in self._vector_dict.items():
            if isinstance(v, torch.Tensor):
                self.__setattr__(k, v.to(device))

    '''magic functions'''

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        from torch.optim.optimizer import Optimizer
        if isinstance(value, torch.device):
            pass
        elif isinstance(value, torch.nn.Module):
            self._model_dict[name] = value
        elif isinstance(value, Optimizer):
            self._optim_dict[name] = value
        elif isinstance(value, (torch.Tensor, np.ndarray)):
            self._vector_dict[name] = value
        elif callable(getattr(value, "state_dict", None)) and callable(getattr(value, "load_state_dict", None)):
            self._other_state_dict[name] = value

    def __setitem__(self, key: str, value: Any):
        self.__setattr__(key, value)

    def train_batch(self, eidx, idx, global_step, batch_data, params: Params, device: torch.device):
        raise NotImplementedError()

    def test_eval_logic(self, dataloader, param: Params):
        raise NotImplementedError()

    def test_eval_logic_v2(self, dataloader, param: Params, is_test: bool):
        return self.test_eval_logic(dataloader, param)

    def predict(self, xs):
        raise NotImplementedError()

    def callbacks(self, params: Params):
        """初始化回调函数"""
        pass

    def datasets(self, params: Params):
        """初始化数据集"""
        pass

    def models(self, params: Params):
        """初始化模型"""
        pass


class Trainer(BaseTrainer):

    def callbacks(self, params: Params):
        pass

    def datasets(self, params: Params):
        pass

    def models(self, params: Params):
        pass

    def train_batch(self, eidx, idx, global_step, batch_data, params: Params, device: torch.device):
        pass

    def extra_state_dict(self) -> dict:
        return super().extra_state_dict()

    def load_extra_state_dict(self, state_dict, strict=False):
        super().load_extra_state_dict(state_dict, strict)


class WrapTrainer(Trainer):

    def __init__(self, params: Params, train_dataloader, model_with_loss_fn, optimize, eval_dataloader=None,
                 test_dataloader=None):
        super().__init__(params)

    def train_batch(self, eidx, idx, global_step, batch_data, params: Params, device: torch.device):
        super().train_batch(eidx, idx, global_step, batch_data, params, device)


class DistributedTrainer():
    def __init__(self, trainer_cls, params: Params, op):
        self.trainer_cls = trainer_cls
        self.params = params
        self.op = op

    def run(self):
        trainer = self.trainer_cls.__new__(self.trainer_cls)  # type:Trainer
        trainer._model_dict = {}  # type:Dict[str,torch.nn.Module]
        trainer._optim_dict = {}  # type:Dict[str,Optimizer]
        trainer._other_state_dict = {}
        trainer._vector_dict = {}
        trainer._checkpoint_plug = {}
        trainer._databundler_dict = {}  # type:Dict[str,DataBundler]
        trainer.train_epoch_toggle = False
        trainer.train_toggle = False
        trainer.experiment = None

        params = self.params
        if self.params is not None:
            trainer.params = params
            if isinstance(params.device, str):
                trainer.regist_device(torch.device(params.device))
            else:
                assert False

            if params.contains('tmp_dir'):
                if params.tmp_dir is not None:
                    os.environ['TMPDIR'] = params.tmp_dir

            if params.local_rank >= 1:
                from thexp import globs
                globs['rank'] = params.local_rank
        else:
            trainer.params = Params()

        import inspect
        from .experiment import Experiment
        # build experiment
        trainer.params.initial()
        file = inspect.getfile(trainer.__class__)
        dirname = os.path.basename(os.path.dirname(file))

        pre = os.path.splitext(os.path.basename(file))[0]

        if not trainer.params.get('git_commit', True):
            os.environ[_OS_ENV.THEXP_COMMIT_DISABLE] = '1'

        trainer.experiment = Experiment("{}.{}".format(pre, dirname))

        # rigist and save params of this training procedure
        trainer.experiment.add_params(params)

        # regist trainer info
        trainer_kwargs = {
            _PLUGIN_KEY.TRAINER.path: inspect.getfile(trainer.__class__),
            _PLUGIN_KEY.TRAINER.doc: trainer.__class__.__doc__,
            _PLUGIN_KEY.TRAINER.fn: pre,
            _PLUGIN_KEY.TRAINER.class_name: trainer.__class__.__name__
        }
        trainer.experiment.add_plugin(_BUILTIN_PLUGIN.trainer, trainer_kwargs)

        params.distributed = True
        import torch.multiprocessing as mp
        if params.world_size == -1:
            params.world_size = torch.cuda.device_count()

        mp.spawn(mp_agent,
                 args=(trainer, self.op),
                 nprocs=trainer.params.world_size)



from thexp import RndManager

import random
import torch.nn as nn
import torch
from torch.optim import SGD


from thexp import RndManager


def test_rndmanager():
    rnd = RndManager()
    lis = []
    rnd_lis = []

    for i in range(2):
        rnd.mark("train")
        data = torch.rand(5, 2)
        y = torch.tensor([0, 0, 0, 0, 0])
        model = nn.Linear(2, 2)

        sgd = SGD(model.parameters(), lr=0.01)
        logits = model(data)
        loss = nn.CrossEntropyLoss()(logits, y)
        loss.backward()
        sgd.step()
        sgd.zero_grad()

        lis.append(list(model.parameters()))

        rnd.shuffle() # 默认设置的种子是time.time()的小数部份，理论上重复概率极低
        rnd_lis.append(random.random())

    a,b = lis
    for pa,pb in zip(a,b):
        assert (pa == pb).all()

    a,b = rnd_lis
    assert a != b


"""

"""
import re
from thexp.decorators import deprecated, warn
from thexp import __VERSION__
import torch
from torch._six import container_abcs, string_classes, int_classes

np_str_obj_array_pattern = re.compile(r'[SaUO]')
default_collate_err_msg_format = (
    "default_collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}")


@deprecated(deprecated_in='1.4.0.10', removed_in='1.5',
            details='This class allocates device during fetching batch of data, may cause memory leak.'
                    'Now `Trainer`/`DataBundler` in `thexp.frame` have added `to(device)` function, which can auto '
                    'allocate device and replace this class.')
class AutoCollate():
    def __init__(self, device):
        super().__init__()
        warn('', DeprecationWarning)
        self.device = device

    def to(self, device):
        self.device = device

    def __call__(self, batch):
        r"""Puts each data field into a tensor with outer dimension batch size"""

        elem = batch[0]
        elem_type = type(elem)
        if isinstance(elem, torch.Tensor):
            out = None
            if torch.utils.data.get_worker_info() is not None:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum([x.numel() for x in batch])
                storage = elem.storage()._new_shared(numel)
                out = elem.new(storage)
            return torch.stack(batch, 0, out=out).to(self.device)
        elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
                and elem_type.__name__ != 'string_':
            elem = batch[0]
            if elem_type.__name__ == 'ndarray':
                # array of string classes and object
                if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                    raise TypeError(default_collate_err_msg_format.format(elem.dtype))

                return self([torch.as_tensor(b) for b in batch])
            elif elem.shape == ():  # scalars
                return torch.as_tensor(batch).to(self.device)
        elif isinstance(elem, float):
            return torch.tensor(batch, dtype=torch.float64)
        elif isinstance(elem, int_classes):
            return torch.tensor(batch).to(self.device)
        elif isinstance(elem, string_classes):
            return batch
        elif isinstance(elem, container_abcs.Mapping):
            return {key: self([d[key] for d in batch]) for key in elem}
        elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
            return elem_type(*(self(samples) for samples in zip(*batch)))
        elif isinstance(elem, container_abcs.Sequence):
            transposed = zip(*batch)
            return [self(samples) for samples in transposed]

        raise TypeError(default_collate_err_msg_format.format(elem_type))


